//
// UNITE jacques' operating system
// (C) 1993 Mattheij's Computer Services
//
// this should become a small real time operating system
// having its ipc ideas and services modeled after qnx. 
// this one will use 386 protected mode though.
//

//
// wait and kill
// both the wait and the kill call are a little hard to implement over a 
// network the way they work now. this probably needs some looking in to
// and changing them into library routines combined with system calls.
//

#include <unite/types.h>
#include <unite/task.h>
#include <unite/config.h>
#include <aout.h>
#include <signal.h>
#include <unite/ports.h>

#include "../include/486.h"
#include "../include/gdt.h"
#include "../include/idt.h"
#include "../include/ldt.h"
#include "../include/sysopc.h"
#include "../include/vmalloc.h"

/*
#define VISUAL_STATES
*/
#if 1
#define INLINE static inline
#define REGISTER register
#else
#define INLINE
#define REGISTER
#endif
/*
#define READY_LIST_ROBUST
*/

/*
#define DEBUG
*/

#if 0
#define TRACE { print(0,2,"       ");print_num(0,2,__LINE__); }
#else
#define TRACE {}
#endif

#define TRUE  1
#define FALSE 0
#define ERROR 0xffffffff
#define NULL  0

// the maximum number of parameters you may pass to a system call
// should be enough for the next 50 years or so

#define MAX_PARAMETERS 20
#define MAX_OBITUARIES 2048

#include "task.h"

PORT ports[PORT_MAX];

//
// obituary lists. every task that dies creates a pack of obituraries.
// one for each admin task. since admins run at a higher priority than
// tasks which die (user programs) we should never run out of obituaries.
// whenever an admin blocks for receive first check its list of pending
// obituaries. if there are any send the first one and put it back on
// the free list.
//


OBITUARY obituaries[MAX_OBITUARIES];
OBITUARY * blank_obituaries;



TASK_INFO task_info[TASK_MAX];									// task info table; one slot for each possible task

unsigned char dft_fp_state[FP_STATE_SIZE];						// default fp state
unsigned char * fp_state;
int *           fp_stored;										// pointer to the flag that remembers where the fp data resides (in the processor or in the struct)
int *           fp_used;										// pointer to the flag that remembers where the fp data resides (in the processor or in the struct)
unsigned char   fp_buf[FP_STATE_SIZE];							// intermediate buffer to restore floating point
																// processor contents from because the 486 can only save and restore
																// fp data from a buffer at a specfic address (pretty stupid)
//
// the ready lists
// 
TASK_INFO * ready_tasks[PRI_MAX];								// pointers to those tasks that are ready to execute
TASK_INFO * ready_tails[PRI_MAX];								// pointers to the last task on each ready list

TASK_INFO * current_task;										// pointer to currently executing task

//
// interrupt vector management data
//
int attached_interrupts[256];									// interrupts attached to by tasks. default is none
int attached_igates[256];
word32 attached_handlers[256];
int stubsize;

//
// speedup variable which eliminates a sequential search from
// the beginning of the ready list each time the next ready task
// at maximum priority is needed. it contains the slot of the highest
// priority list last written
//
int current_priority = 1;										// start searching from this priority onwards

//
// dont change the definitions below without updating 
// the boot loader
//

#define MAIN_TSS_ID 1

#define STACKSIZE 1024

#define P_ISTACK_SS     interrupt_stack[(STACKSIZE - 2)]
#define P_ISTACK_ESP    interrupt_stack[(STACKSIZE - 3)]
#define P_ISTACK_EFLAGS interrupt_stack[(STACKSIZE - 4)]
#define P_ISTACK_CS     interrupt_stack[(STACKSIZE - 5)]
#define P_ISTACK_EIP    interrupt_stack[(STACKSIZE - 6)]
#define P_ISTACK_ERROR  interrupt_stack[(STACKSIZE - 7)]

#define N_ISTACK_EFLAGS interrupt_stack[(STACKSIZE - 2)]
#define N_ISTACK_CS     interrupt_stack[(STACKSIZE - 3)]
#define N_ISTACK_EIP    interrupt_stack[(STACKSIZE - 4)]
#define N_ISTACK_ERROR  interrupt_stack[(STACKSIZE - 5)]

#define MAX_BOOT_TASKS 15

typedef struct {
	word16 cseg;				/* real mode code segment for task */
	word16 dseg;				/* real mode data segment for task */
	word32 taddr;				/* lineair text address */
	word32 daddr;				/* lineair data address */
	word16 priority;
	word16 filler;
	char name[0x10];			/* name of the task */
	GNU_AOUT header;			/* used to load a.out files */
	word32 argc;
	word32 argv;
	word32 envp;
} BOOT_TASK;

extern int n_boot_tasks;
extern BOOT_TASK boot_tasks[MAX_BOOT_TASKS];
extern TSS boot_tss[2];

extern DESCRIPTOR gdt[GDT_MAX];
extern GATE       idt[IDT_MAX];

word32 interrupt_stack[STACKSIZE];

//
// some 386 exceptions push an error word.
// it will end up here
//

word32 error_word;

extern "C" {
	void ivec0(void);
	void ivec1(void);
	void timer_interrupt(void);
	void page_fault_handler(void);
};

extern int int_number;
extern int hw_port_signalled;
extern unsigned char hw_ports_signalled[1024];

// the amount of physical memory

extern word32 physical_memory;

// how many bytes available between end of stack and 640k

word32 space_below_640;

// where sits the data segment of the tasker in physical memory
// (used to find pages in vmalloc.cc)

word32 __daddr;

// __end is used to be able to reserve the memory occupied by the kernel.
// this has to be claimed before calling vmem_init because otherwise the
// first page table will be overlayed with the interrupt vectors
// (this is not the intention).

word32 cr3;

extern void * itmemcpy(word16 s,word32 d,word32 s,word32 l);

extern "C" {
	int inportb(int);
	void outportb(int,int);
	void fprestore(void);
	void fp_restore(char *);
	void fpsave(char *);
	void nosignal_hw_port(void);
	void signal_hw_port(void);
	void _disable(void);
	void _enable(void);
};

//
// parameter to sys_exit
//

#define TRANSFORMING 1
#define NOT_TRANSFORMING 0

//
// timer definitions
//

#define SLICE       1000										// slices per second
#define TIMER0      0x40										// timer 0 port
#define TIMER1      0x41										// timer 1 port
#define TIMER2      0x42										// timer 2 port
#define TIMER_MODE  0x43										// timer mode register
#define SQUARE_WAVE 0x36										// continuous square wave output for timer 0
#define TIMER_FREQ 1193182										// timer base frequency in Hz
#define TIMER_COUNT (TIMER_FREQ / SLICE)						// value to load into counter 0 

#define TASK_ALIVE(t) (((TASK_INFO*)t)->state != TASK_STATE_FREE && ((TASK_INFO*)t)->state != TASK_STATE_ZOMBIE && ((TASK_INFO*)t)->state != TASK_STATE_DUMPED)

/*
 * possibilities for default signals
 */

#define CORE    1
#define KILL    2
#define DISCARD 3

static int sigdefaults[NSIG] = {
	0,			/* 0 */
	KILL,		// SIGHUP	1	/* hangup */
	KILL,		// SIGINT	2	/* interrupt (rubout) */
	CORE,		// SIGQUIT	3	/* quit (ASCII FS) */
	CORE,		// SIGILL	4	/* illegal instruction (not reset when caught) */
	CORE,		// SIGTRAP	5	/* trace trap (not reset when caught) */
	CORE,		// SIGIOT	6	/* IOT instruction */
				// SIGABRT  6	/* used by abort, replace SIGIOT in the future */
	CORE,		// SIGEMT	7	/* EMT instruction */
	CORE,		// SIGFPE	8	/* floating point exception */
	KILL,		// SIGKILL	9	/* kill (cannot be caught or ignored) */
	CORE,		// SIGBUS	10	/* bus error */
	CORE,		// SIGSEGV	11	/* segmentation violation */
	CORE,		// SIGSYS	12	/* bad argument to system call */
	KILL,		// SIGPIPE	13	/* write on a pipe with no one to read it */
	KILL,		// SIGALRM	14	/* alarm clock */
	KILL,		// SIGTERM	15	/* software termination signal from kill */
	KILL,		// SIGUSR1	16	/* user defined signal 1 */
	KILL,		// SIGUSR2	17	/* user defined signal 2 */
	DISCARD,	// SIGCLD	18	/* child status change */
				// SIGCHLD	18	/* child status change alias (POSIX) */
	DISCARD,	// SIGPWR	19	/* power-fail restart */
	DISCARD,	// SIGWINCH 20	/* window size change */
	DISCARD,	// SIGURG	21	/* urgent socket condition */
	DISCARD,	// SIGPOLL  22	/* pollable event occured */
				// SIGIO	SIGPOLL	/* socket I/O possible (SIGPOLL alias) */
	DISCARD,	// SIGSTOP 23	/* stop (cannot be caught or ignored) */
	DISCARD,	// SIGTSTP 24	/* user stop requested from tty */
	DISCARD,	// SIGCONT 25	/* stopped process has been continued */
	DISCARD,	// SIGTTIN 26	/* background tty read attempted */
	DISCARD,	// SIGTTOU 27	/* background tty write attempted */
	DISCARD,	// SIGVTALRM 28	/* virtual timer expired */
	DISCARD,	// SIGPROF 29	/* profiling timer expired */
	DISCARD,	// SIGXCPU 30	/* exceeded cpu limit */
	DISCARD,	// SIGXFSZ 31	/* exceeded file size limit */
};

// forward definition of signal_pending so sys_replym may call it if a
// task had signals locked during a reply.

void signal_pending(TASK_INFO * tp);


// forward definition of sys_kill so sys_kill(parent,SIGCHLD) may be sent
// from exit of a task

int sys_kill(int tid,int signal);

void
sys_select_unblock(int tid,int why);

SYSTEM_TIMERS system_timers;

TSS hardware_interrupt_tss[16];
TSS page_fault_handler_tss;

unsigned char page_fault_handler_stack[1024];

#define _disable() asm("cli")
#define _enable()  asm("sti")

// the kernel code segments; one for each possible code segment
// with the number of users at the moment.
// if the number of users is 0 the code segment is not used
// if the number of users is 1 the code segment is ready for re-use
// if the number of users >  1 the code segment is active and may not
// be used.

typedef struct {
	word32 vbase;
	int    nusers;
} K_CODE_SEGMENT;

K_CODE_SEGMENT code_segments[TASK_MAX];

int
sys_free_old_code_segment()

{
	int i;

	for (i=0;i<TASK_MAX;i++) {
		if (code_segments[i].nusers == 1) {
			vmem_free(code_segments[i].vbase);
			code_segments[i].vbase  = 0;
			code_segments[i].nusers = 0;
			// ok this one freed try again in caller
			return 0;
		}
	}
	// nothing to free
	return -1;
}

//
// change a segment/offset pair into a physical 32 bit address
//

word32
phys(word16 s,word16 o)

{
	word32 r;

	r = s;

	r <<= 4;

	r += o;

	return r;
}

void
internal_memset(unsigned char * t,int v,int l)

{
	while (l--) {
		*t++ = v;
	}
}


void
internal_memcpy(unsigned char * d,unsigned char * s,int l)

{
	while (l--) {
		*d++ = *s++;
	}
}

//
// routine used to fool the optimizer into leaving empty loops in place
// used to slow things down so you can see what's going on
//

void
nothing()
{
}

char * messages[] = {
	"Divide Error",
	"Debugger Call",
	"NMI Interrupt",
	"Breakpoint",
	"INTO-detected Overflow",
	"BOUND Range Exceeded",
	"Invalid Opcode",
	"Device Not Available",
	"Double Fault",
	"Intel Reserved",
	"Invalid Task State Segment",
	"Segment Not Present",
	"Stack Exception",
	"General Protection",
	"Page Fault",
	"Intel Reserved 2",
	"Floating-Point Error",
	"Alignment Check"
};
void
vid_mem_spinner(int x,int y)

{
	asm("	push	%es        		");
	asm("	movw	$0x88,%eax		");
	asm("	movw	%ax,%es			");
	asm("	movl	12(%ebp),%eax	");
	asm("	imul	$80,%eax		");
	asm("	addw	 8(%ebp),%eax	");
	asm("	addl	%eax,%eax		");
	asm("	es						");
	asm("	addb	$1,(%eax)		");
	asm("	incl	%eax			");
	asm("	movw	$0x07,%dl		");
	asm("	es						");
	asm("	movb	%dl,(%eax)		");
	asm("	pop		%es             ");
}


void
vid_mem_putchar(int x,int y,int c)

{
	asm("	push	%es        		");
	asm("	movw	$0x88,%eax		");
	asm("	movw	%ax,%es			");
	asm("	movl	12(%ebp),%eax	");
	asm("	imul	$80,%eax		");
	asm("	addw	 8(%ebp),%eax	");
	asm("	addl	%eax,%eax		");
	asm("	movw	16(%ebp),%dx	");
	asm("	es						");
	asm("	movb	%dl,(%eax)		");
	asm("	incl	%eax			");
	asm("	movw	$0x07,%dl		");
	asm("	es						");
	asm("	movb	%dl,(%eax)		");
	asm("	pop		%es             ");
}

void
put_nibble(int x,int y,int n)

{
	n &= 0x0f;

	if (n > 9) {
		n += 'A' - 10;
	}
	else {
		n += '0';
	}
	vid_mem_putchar(x,y,n);
}

void
dump_bytes(int x,int y,unsigned char * s,int n)

{
	while (y > 59) {
		y -= 60;
	}
	while (n--) {
		put_nibble(x++,y,*s >> 4);		
		put_nibble(x++,y,*s);
		vid_mem_putchar(x++,y,' ');
		s++;
	}
}

void
print(int x,int y,char * s)

{
	while (*s) {
		vid_mem_putchar(x++,y,*s++);
	}
}

void
printv(int x,int y,char * s)

{
	while (*s) {
		vid_mem_putchar(x,y++,*s++);
	}
}

int
print_num(int x,int y,int n)

{
	int d;

	if (n == 0) {
		return x;
	}
	
	d = n % 10;
	n /= 10;
	x = print_num(x,y,n);

	vid_mem_putchar(x,y,d + '0');
	x++;
	return x;
}

int
print_numv(int x,int y,int n)

{
	int d;

	if (n == 0) {
		return y;
	}
	
	d = n % 10;
	n /= 10;
	y = print_numv(x,y,n);

	vid_mem_putchar(x,y,d + '0');
	y++;
	return y;
}

char tss_used[104] = {
	1,1,0,0,1,1,1,1,
	1,1,0,0,1,1,1,1,
	1,1,0,0,1,1,1,1,
	1,1,0,0,1,1,1,1,
	1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,
	1,1,0,0,1,1,0,0,
	1,1,0,0,1,1,0,0,
	1,1,0,0,1,1,0,0,
	1,1,0,0,1,1,1,1 
};
void
dump_tss(TSS * t)

{
	unsigned char * p;
	int i;
	int x;
	int y;

	print(0, 0,"        -- -- xx xx ldt       ");
	print(0, 1,"        -- -- xx xx gs        ");
	print(0, 2,"        -- -- xx xx fs        ");
	print(0, 3,"        -- -- xx xx ds        ");
	print(0, 4,"        -- -- xx xx ss        ");
	print(0, 5,"        -- -- xx xx cs        ");
	print(0, 6,"        -- -- xx xx es        ");
	print(0, 7,"        xx xx xx xx edi       ");
	print(0, 8,"        xx xx xx xx esi       ");
	print(0, 9,"        xx xx xx xx ebp       ");
	print(0,10,"        xx xx xx xx esp       ");
	print(0,11,"        xx xx xx xx ebx       ");
	print(0,12,"        xx xx xx xx edx       ");
	print(0,13,"        xx xx xx xx ecx       ");
	print(0,14,"        xx xx xx xx eax       ");
	print(0,15,"        xx xx xx xx eflags    ");
	print(0,16,"        xx xx xx xx eip       ");
	print(0,17,"        xx xx xx xx cr3 (pdbr)");
	print(0,18,"        -- -- xx xx ss2       ");
	print(0,19,"        xx xx xx xx esp2      ");
	print(0,20,"        -- -- xx xx ss1       ");
	print(0,21,"        xx xx xx xx esp1      ");
	print(0,22,"        -- -- xx xx ss0       ");
	print(0,23,"        xx xx xx xx esp0      ");
	print(0,24,"        -- -- xx xx link      ");

	i = 0;
	p = (unsigned char*) t;
	for (y=24;y>=0;y--) {
		for (x=0;x<4;x++) {
			if (tss_used[i]) {
				dump_bytes(17 - 3 * x,y,p,1);
			}
			i++;
			p++;
		}
	}
}

void
show_task_states()

{
	int i;

	for (i=0;i<80;i++) {
		vid_mem_putchar(i,24,task_info[i].state + '0');
		printv(i,18,"   ");
		if (task_info[i].blocked_on != -1) {
			print_numv(i,18,task_info[i].blocked_on);
		}
	}
}

void
dump_stack(unsigned char * start,unsigned char * end)

{
	unsigned char bytes[4];
	int y;
	int t;

	y = 24;

	while (y >= 0 && start <= (end-4)) {
		itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)bytes,(word32) start,4);
		start += 4;
		t = bytes[0];		
		bytes[0] = bytes[3];
		bytes[3] = t;
		t = bytes[2];
		bytes[2] = bytes[1];
		bytes[1] = t;
		dump_bytes(45,y--,bytes,4);
	}
}

//
// set an idt entry to point to its stub
// or possibly to a handler
//

void
init_gate(GATE * gate,word16 sel,word32 offset,word8 type)

{
	gate->selector = sel;
	gate->offset_lo = (word16)(offset & 0xffff);
	gate->offset_hi = (word16)(offset >> 16);
	gate->type_dpl = type;
	gate->count    = 0;
}

//
// set up a descriptor in the gdt
//

void
init_descriptor(int i,word32 base,word32 limit,word8 type,word16 operation_size)

{
	DESCRIPTOR * descr;

	if ((i >> 3) >= GDT_MAX) {
		return;
	}

	descr = &gdt[i >> 3];

	descr->base_lo  = (word16) base;
	descr->base_mid = (word8) (base >> 16);
	descr->type_dpl = type;

	if (limit > 0xfffffL) {
		limit >>= 12;
		descr->limit_hi = ((word8)(limit >> 16) & 0x0f) | SEG_GRANULARITY;
	}
	else {
		descr->limit_hi = ((word8)(limit >> 16) & 0x0f);
	}

	descr->limit_hi |= operation_size;

	descr->limit_lo = (word16)limit;
	descr->base_hi = (word8)(base >> 24);

//	dump_bytes(10,i/8,(char*) descr,8);
}

//
// fill a task state segment
//

void
set_interrupt_stacks(TSS * t)

{
#if 1
	t->esp0 = (word32)&interrupt_stack[STACKSIZE - 1];
	t->ss0  = TASK_DSEL;

	t->esp1 = (word32)&interrupt_stack[STACKSIZE - 1];
	t->ss1  = TASK_DSEL;

	t->esp2 = (word32)&interrupt_stack[STACKSIZE - 1];
	t->ss2  = TASK_DSEL;
#else
	t->esp0 = (word32)&interrupt_stack[STACKSIZE - 1];
	t->ss0  = INTS_DSEL;

	t->esp1 = (word32)&interrupt_stack[STACKSIZE - 1];
	t->ss1  = INTS_DSEL;

	t->esp2 = (word32)&interrupt_stack[STACKSIZE - 1];
	t->ss2  = INTS_DSEL;
#endif

	t->esp  &= 0xfffffff0;

#if 0
	t->esp0 &= 0xfffffff0;
	t->esp1 &= 0xfffffff0;
	t->esp2 &= 0xfffffff0;
#endif
}

void
init_tss(TSS * t,int privilege,word16 cs,word16 ds,word32 esp,word32 eax, word32 edi,word32 esi,word32 ebx,word32 eip,word32 daddr,word32 ecx,word32 ebp)

{
	int i;

	internal_memset((unsigned char*)t,0,sizeof(TSS));

	//
	// reset io bits to allow acces to all io ports
	//
	t->iomap = (char*)&t->iobits -  (char*)t;

	for (i=0;i<64;i++) {
		t->iobits[i] = 0x0000;
	}

	t->cs = cs | privilege;
	t->ds = ds | privilege;
	t->es = ds | privilege;
	t->ss = ds | privilege;
	t->fs = ds | privilege;
	t->gs = ds | privilege;
	t->eip = eip;
	t->esp = esp;

	// some parameters to the newly hatched programs

	t->edi = edi;	// argc
	t->esi = esi;	// argv
	t->ebx = ebx;	// envp
	t->eax = eax;	// pid
	t->edx = daddr;	// daddr (start of data segment in virtual memory)
	t->ecx = ecx;   // fddata in case of execve
	t->ebp = ebp;	// fdnbytes in case of execve

	t->cr3 = cr3;

	set_interrupt_stacks(t);
}

//
// a task has crashed or done something inexcusable
//

void
core_dump(void)

{
}

//
// add a task to the tail of the ready list for its priority.
//

INLINE void
add_to_ready_list(REGISTER TASK_INFO * tp)

{
	REGISTER int priority;

#ifdef READY_LIST_ROBUST
	if (tp->in_list) {
		return;
	}

	tp->in_list = TRUE;
#endif

	priority = tp->priority;

	if (priority < current_priority) {
		current_priority = priority;
	}

	if (ready_tasks[priority]) {
		//
		// list already has tasks in it; add the new one to the tail of the list
		// 
		ready_tails[priority]->next_ready = tp;
		ready_tails[priority] = tp;
	}
	else {
		//
		// list is empty; create a new one. initial tail is the same as start
		//
		ready_tasks[priority] = tp;
		ready_tails[priority] = tp;
	}

	tp->state = TASK_STATE_READY;
	tp->next_ready = NULL;										// this will be the last entry in the list
	tp->blocked_on = -1;
}

//
// copy data from a message list in one task to a message list in another
// possibly starting at an offset.
//

int
itlistcpy(TASK_INFO * dt,K_MSG_LIST * dlist,TASK_INFO * st,K_MSG_LIST * slist)

{
//	static int t = 10;
	int tn;
	int n;			// number of bytes to copy

	tn = 0;

	// offset the message lists with the base address of the task

	dlist->list += dt->daddr;
	slist->list += st->daddr;

	while (TRUE) {
		if (dlist->left == 0) {
			dlist->index++;
			if (dlist->index == dlist->count) {
				break;
			}
			itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)&dlist->buf,dlist->list+8*dlist->index,8);
//			dump_bytes(0,t++,(char*)&dlist->buf,8);
			dlist->buf += dt->daddr;
		}
		if (slist->left == 0) {
			slist->index++;
			if (slist->index == slist->count) {
				break;
			}
			itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)&slist->buf,slist->list+8*slist->index,8);
//			dump_bytes(1,t++,(char*)&slist->buf,8);
			slist->buf += st->daddr;
		}

		if (slist->left > dlist->left) {
			n = dlist->left;
		}
		else {
			n = slist->left;
		}

		itmemcpy(LINM_DSEL,dlist->buf,slist->buf,n);
		tn += n;

		dlist->left -= n;
		slist->left -= n;
		dlist->buf  += n;
		slist->buf  += n;
	}

	return tn;
}

/* INLINE */

int
drop_obituary(TASK_INFO * to)

{
	OBITUARY * op;
	op = to->obituaries;
	// and mark the obituary as processed anyway. (so the task slot may be freed)
	to->obituaries = (OBITUARY *) op->next;
	op->next = blank_obituaries;
	blank_obituaries = op;
	return op->tid | TID_DEAD;
}

void
wakeup_recv(REGISTER TASK_INFO * to)

{
	int i;
	/* BUG ! REGISTER */ TASK_INFO * from;

	if (to->ipcunblock && to->state == TASK_STATE_SELECT) {
		//
		// this task is select blocked but also waiting for
		// ipc events. check if any ipc events are pending and 
		// if neccessary unblock the task.
		//

		//
		// check if there is a signal on a port pending for this task.
		// this is done with a list for speed.
		// the list hangs from the task structure and contains all
		// the ports that were signalled
		//
		
		if (to->ports_signalled) {
			i = to->ports_signalled->num;
			ports[i].val--;
			if (ports[i].val == 0) {
				to->ports_signalled = (PORT *)ports[i].next;
				ports[i].next = NULL;
			}
			sys_select_unblock(to->tid,i | TID_PORT);
			return;
		}
	
		//
		// check for pending obituaries (only for admins)
		// only send them if the admin is doing a general receive 
		// which hopefully sits in the main loop. if not this obituary will be lost
		//
	
		if (to->obituaries) {
			sys_select_unblock(to->tid,drop_obituary(to));
			return;
		}
	
		from = (TASK_INFO*) to->tasks_send_blocked;
		if (from != NULL) {
			sys_select_unblock(to->tid,from->tid | TID_TASK);
			return;
		}															// nobody send blocked; forget it
	}
	
	if (to->state != TASK_STATE_RECV) {
//		printf("task %d not receiving\n",to->tid);
		return;
	}

	//
	// check if there is a signal on a port pending for this task.
	// this is done with a list for speed.
	// the list hangs from the task structure and contains all
	// the ports that were signalled
	//

	if ((to->blocked_on == -1) || (to->blocked_on & TID_MASK) == TID_PORT) {
		// needs a check on the right port number !
		// WERK
		if (to->ports_signalled) {
			i = to->ports_signalled->num;
			ports[i].val--;
			if (ports[i].val == 0) {
				to->ports_signalled = (PORT *)ports[i].next;
				ports[i].next = NULL;
			}
			to->tss.eax = i | TID_PORT;
			add_to_ready_list(to);
//			vid_mem_putchar(20,20+i,'0'+ports[i].val);
			return;
		}
		if ((to->blocked_on & TID_MASK) == TID_PORT) {
			return;
		}
	}

	//
	// check for pending obituaries (only for admins)
	// only send them if the admin is doing a general receive 
	// which hopefully sits in the main loop. if not this obituary will be lost
	//

	if (to->blocked_on == -1 && to->obituaries) {
		to->tss.eax = drop_obituary(to);
		add_to_ready_list(to);
		return;
	}

	from = (TASK_INFO*) to->tasks_send_blocked;
	if (from == NULL) {
//		printf("task %d has no tasks send blocked on it\n",to->tid);
		return;
	}															// nobody send blocked; forget it

	if (from->state != TASK_STATE_SEND) {
//		printf("oops task in tasks_send_block list but not send blocked\n");
		return;
	}

	//
	// here should be a test to see if the receiving task did a specific
	// request for data from a certain task. for now assume that any non-general
	// receive was done to get a signal from a port (await)
	//

	if (to->blocked_on != -1 && from->tid != to->blocked_on) {
		// task has tasks send blocked but it requested something else
		// (lets hope it waits for an interrupt)
		return;
	}

	//
	// remove the sending task from the send blocked on list
	//
//	printf("transferring message from task %d to task %d\n",from->tid,to->tid);

	to->tasks_send_blocked = from->next_send_blocked;
	from->next_send_blocked = NULL;								

	to->receive_list.index = -1;
	to->receive_list.left  = 0;
	from->send_list.index  = -1;
	from->send_list.left   = 0;

	itlistcpy(to,&to->receive_list,from,&from->send_list);

	from->state = TASK_STATE_REPLY;								// sending task is now reply blocked
	from->reply_locked = TRUE;									// assume this will be a short session.
																// if it is not the admin will reply-unlock this task.

	add_to_ready_list(to);										// receiving task is ready

	to->tss.eax = from->tid;									// keep tabs on where the message came from
}

//
// system calls. (send receive & reply)
// a system call is executed on the stack of the
// call that issues it. every system call ends in a
// timeslice
//

//
// this function will send a message into the memory space of another
// task the moment that task issues a receive. if the task is already
// receive blocked the transfer is instantaneous. the sending task will
// be unblocked at the reply from the receiving task.
//

INLINE void
sys_sendm(REGISTER TASK_INFO * tp,int tid,int scount,int rcount,word32 slist,word32 rlist)

{
	REGISTER TASK_INFO * dt;

	dt = &task_info[tid];

	if (dt->state == TASK_STATE_FREE) {
		tp->tss.eax = ERROR;
		return;
	}

	tp->blocked_on       = tid;
	tp->send_list.list   = slist;	
	tp->reply_list.list  = rlist;
	tp->send_list.count  = scount;
	tp->reply_list.count = rcount;
	//
	// // put this task at the head of the send blocked on me list of the
	// // destination task
	// // it seems unfair to put the last task to send a message to another one
	// // at the beginning of the send blocked list for the destination task;
	// // however there are two good reasons to do this:
	// //   1) it is faster
	// //   2) it makes sure that if the destination task did a specific receive
	// //      from another task and the send blocked queue is not empty that
	// //      every task sending to it afterwards gets checked. if we were to
	// //      put new entries at the back of the list wakeup_recv would
	// //      only check the first entry in the list and miss the one being waited for
	// 
	// no longer... from now on the order of serving clients is the order of
	// appearance. putting entries at the head of the list caused a lot
	// of trouble if the admin had a lot more work than the clients to generate
	// the data. 
	//
	if (dt->tasks_send_blocked == NULL) {
		// task is the first task to become blocked on destination.
		dt->tasks_send_blocked = tp;
	}
	else {
		TASK_INFO * tt;
		// task is not the first task to become send blocked on destination
		// find the end of the list
		tt = dt->tasks_send_blocked;
		while (tt->next_send_blocked) {
			tt = tt->next_send_blocked;
		}
		// and attach then sending task
		tt->next_send_blocked = tp;
	}
	tp->next_send_blocked = NULL;
	//
	// task is now officially send blocked
	//
	tp->state = TASK_STATE_SEND;
	//
	// attempt to wake up the destination task
	//
	wakeup_recv(dt);
	//
	// assume the send will succeed and already fill in the dest tid
	//
	tp->tss.eax = tid;
}

//
// receive receives a message from an external source and
// places it in the receive_data buffer. receive_len 
// indicates the size of the buffer. the function returns
// the fd of the sender
// even if the task wants to do an exclusive receive from another
// task and it has a signal pending the signal will get preference
// I''m not sure if this is right
//

INLINE void
sys_receivem(REGISTER TASK_INFO * tp,int tid,int rcount,word32 rlist)

{
	tp->tss.eax = 0xffffffff;

	if ((tid & TID_MASK) == 0 || (tid & TID_MASK) == TID_TASK) {
		if (tid < 0 || tid >= TASK_MAX) {
			/* invalid task id */
			return;
		}
		if (task_info[tid].state == TASK_STATE_FREE) {
			/* task slot not in use */
			return;			
		}
	}
	tp->receive_list.list  = rlist;
	tp->receive_list.count = rcount;
	tp->state              = TASK_STATE_RECV;	
	tp->blocked_on         = tid;

//	printf("suspending task %d on receive\n",tp->tid);
	//
	// try to wake up the task 
	//
	wakeup_recv(tp);
}

//
// reply replies to a task which is waiting for a reply
// after a blocking send. reply returns the tid if the transfer
// went ok; -1 if the task wasn't waiting for a reply or if it
// died.
//

INLINE void
sys_replym(REGISTER TASK_INFO * tp,int tid, int rcount, word32 rlist)

{
	REGISTER TASK_INFO * dt;

	tp->reply_list.list  = rlist;
	tp->reply_list.count = rcount;

	dt = &task_info[tid];

	if (dt->state != TASK_STATE_REPLY || dt->blocked_on != tp->tid) {
		//
		// the task woke up or has gone to heaven
		// or is blocked on another task
		//
		tp->tss.eax = ERROR;
		return;
	}

	dt->reply_list.index = -1;
	dt->reply_list.left  = 0;
	tp->reply_list.index = -1;
	tp->reply_list.left  = 0;

	itlistcpy(dt,&dt->reply_list,tp,&tp->reply_list);

	if (dt->reply_locked) {	
		// this task has been reply locked for a while. check if it has
		// signals pending.
		signal_pending(dt);
		dt->reply_locked = FALSE;
	}

	if (dt->state != TASK_STATE_READY) {
		// signal_pending did not make the task ready. do it anyway.
		// (this test to avoid putting dt in the ready list twice)
		add_to_ready_list(dt);
	}

	tp->tss.eax = tid;
}

INLINE int
sys_signal_port(int port)

{
//	print(10,10,"in sys signal port");

	if (port > 0 && port < PORT_MAX) {
		if (ports[port].tid == -1) {
			return ERROR;
		}														// nobody attached to this port; this must be an error

		ports[port].val++;										// increment semaphore instead of simply setting to avoid losing interrupts
		//
		// if this is the signal that changes this ports state
		// from 0 to 1 then add the port to the ports signalled list
		if (ports[port].val == 1) {
#if 0
			// adding to the front is not very fair; but it's much faster
			// than adding to the end. if the new port state is not equal to
			// 1 the port is already in the list
			//
			ports[port].next = task_info[ports[port].tid].ports_signalled;
			task_info[ports[port].tid].ports_signalled = &ports[port];
#else
			// its also wrong. now we add at the end. probably there are
			// not a lot of entries in the list anyway and if it is always			
			// done at the front one entry might get preference
			// like this the 'interrupts' are handled in sequence.
			ports[port].next = NULL;
			if (task_info[ports[port].tid].ports_signalled == NULL) {
				// nothing yet; put the new entry at the front
				task_info[ports[port].tid].ports_signalled = &ports[port];
			}
			else {
				PORT * p;
				p = task_info[ports[port].tid].ports_signalled;
				while (p->next != NULL) {
					p = p->next;
				}
				p->next = &ports[port];
			}

#endif
		}

		wakeup_recv(&task_info[ports[port].tid]);

//		vid_mem_putchar(20,20+port,'0'+ports[port].val);

		return port;
	}
//	print(10,12,"port number error");
	return ERROR;
}

//
// port_attach lets a task attach to a specified port ( < PORT_ALLOC_BASE)
// or allocates one from the pool
// on success the port number is returned; on failure -1
//

INLINE void
sys_port_attach(REGISTER TASK_INFO * tp,int port)

{
	int i;
	
	//
	// depressive person. (assume this will go wrong)
	//
	tp->tss.eax = ERROR;

	if (port != -1) {
		//
		// don't allow specific allocation of ports >= PORT_ALLOC_BASE
		//
		if (port >= PORT_ALLOC_BASE) {
			return;
		}

		if (ports[port].tid == -1) {
			ports[port].tid = tp->tid;
			ports[port].val = 0;
			tp->tss.eax = port;
			return;
		}
		return;
	}

	for (i=PORT_ALLOC_BASE;i<PORT_MAX;i++) {
		if (ports[i].tid == -1) {
			ports[i].tid = tp->tid;
			ports[i].val = 0;
			tp->tss.eax = i;
			return;
		}
	}
}

INLINE void
sys_port_detach(REGISTER TASK_INFO * tp,int port)

{
	if (ports[port].tid == tp->tid) {	
		//
		// if the port still has a value that means it is also still in
		// the ports signalled list
		//
		if (ports[port].val) {
		}
		ports[port].tid = -1;
		tp->tss.eax = port;
		return;
	}

	tp->tss.eax = ERROR;
}

INLINE void
sys_force_ready(REGISTER TASK_INFO * tp,int tid)

{
	tp->tss.eax = ERROR;
	//
	// check for bad task id's
	//
	if (tid < 0 || tid >= TASK_MAX) {
		return;
	}

	//
	// check for tasks that are not active
	//
	if (task_info[tid].state == TASK_STATE_FREE) {
		return;		
	}

	//
	// seems to be ok.
	// ignore when already ready
	//
	if (task_info[tid].state != TASK_STATE_READY) {
		if (task_info[tid].state == TASK_STATE_SELECT) {
			task_info[tid].select_sig_block = TRUE;
		}

		// let the task now that it was forced ready and not through the end of the system call
		task_info[tid].tss.eax = 0xffffffff;
		add_to_ready_list(&task_info[tid]);
	}

	tp->tss.eax = tid;
}

INLINE void
sys_interrupt_attach(REGISTER TASK_INFO * tp,int vector,int port,word32 handler)

{
	tp->tss.eax = ERROR;

	if (vector >= 0 && vector <= 256 && port > 0 && port < PORT_MAX && ports[port].tid == tp->tid) {
		if (attached_interrupts[vector] == -1) {
			attached_interrupts[vector] = tp->tid;
			attached_handlers[vector] = handler;
			tp->tss.eax = vector;
		}
	}
}

INLINE void
sys_igate_attach(REGISTER TASK_INFO * tp,int vector,word32 stack,word32 handler)

{
	int gdt_slot;
	tp->tss.eax = ERROR;

	// check the handler address for validity
	// the first page is not used and the handler address should be smaller than the
	// the size of the text segment

	if (handler < 4096 && handler >= tp->tlen) {
		return;
	}

	if (vector >= 0x70 && vector < 0x80) {
		if (attached_interrupts[vector] == -1) {
			attached_interrupts[vector] = tp->tid;
			attached_igates[vector]     = TRUE;

			gdt_slot = VECT_GSEL + (vector << 3);

			init_descriptor(gdt_slot,task_info[0].daddr + (word32)&hardware_interrupt_tss[vector - 0x70],TSS_SIZE-1,TYPE_TSS_DESCR | TYPE_DPL_3,0);
			init_descriptor(gdt_slot+(48 << 3),task_info[0].daddr + (word32)&hardware_interrupt_tss[vector - 0x70],TSS_SIZE-1,TYPE_DATA_DESCR | TYPE_DPL_0,SEG_DEFAULT);

			internal_memcpy((unsigned char *)&hardware_interrupt_tss[vector - 0x70],(unsigned char *)&tp->tss,sizeof(TSS));

			hardware_interrupt_tss[vector - 0x70].eip = handler;
			hardware_interrupt_tss[vector - 0x70].eflags &= 0xffffbdffl;
			hardware_interrupt_tss[vector - 0x70].esp     = stack;

			// this task attaches an interrupt handler. from now on it is 
			// an integral part of the kernel and will run at permission level 0
			// dpl 0 (so the handler will run allways)

			gdt[(gdt_slot >> 3) + 16] = gdt[tp->tss.cs >> 3];
			gdt[(gdt_slot >> 3) + 32] = gdt[tp->tss.ds >> 3];

			hardware_interrupt_tss[vector - 0x70].cs = gdt_slot + (16 << 3);
			hardware_interrupt_tss[vector - 0x70].ds = gdt_slot + (32 << 3);
			hardware_interrupt_tss[vector - 0x70].es = gdt_slot + (32 << 3);
			hardware_interrupt_tss[vector - 0x70].fs = gdt_slot + (32 << 3);
			hardware_interrupt_tss[vector - 0x70].gs = gdt_slot + (32 << 3);
			hardware_interrupt_tss[vector - 0x70].ss = gdt_slot + (32 << 3);

			gdt[(gdt_slot >> 3) + 16].type_dpl &= ~TYPE_DPL_3;
			gdt[(gdt_slot >> 3) + 32].type_dpl &= ~TYPE_DPL_3;

			init_gate(&idt[vector],gdt_slot,0,TYPE_TASK_GATE | TYPE_DPL_0);

			tp->tss.eax = vector;
		}
	}
}

void
sys_sbrk(TASK_INFO * tp,int incr)

{
	word32 old_brk;

	// werk ! hier aantal paginas toevoegen aan het data segment.

	if (incr > 0) {
		old_brk = tp->brk;

		while (vmem_map((void *)tp->daddr + tp->brk,(word32) incr) < 0) {
			// the map failed; try to free up some more memory
			if (sys_free_old_code_segment() < 0) {
				// there was no more memory to free
				tp->tss.eax = 0xffffffff;
				return;
			}
		}

		tp->brk += incr;

		tp->tss.eax = old_brk;
	}
	else {
		tp->tss.eax = tp->brk;
	}
}

void
wakeup_wait(TASK_INFO * tp)

{
	int seen_children;
	int i;

	if (tp->state != TASK_STATE_WAIT) {
		// task is not waiting so forget about it.
		return;
	}

	// task is waiting. now find any zombie children.
	// so far no children seen
	seen_children = FALSE;

	for (i=0;i<TASK_MAX;i++) {
		if (task_info[i].parent_tid == tp->tid) {
			if (TASK_ALIVE(&task_info[i]) || task_info[i].state == TASK_STATE_ZOMBIE) {
				seen_children = TRUE;
			}
			if (task_info[i].state == TASK_STATE_ZOMBIE) {
				// found one !.
				// move pid and exit status and pid to parent
				tp->tss.eax = i;
				if (tp->wait_status_pointer != NULL) {
					itmemcpy(LINM_DSEL,tp->daddr+tp->wait_status_pointer,task_info[0].daddr + (word32) &task_info[i].exit_status,sizeof(int));
				}
				// release task slot
				task_info[i].state = TASK_STATE_FREE;
				// and ready the parent task.
				add_to_ready_list(tp);
				// only one task per wait.
				return;
			}
		}
	}

	if (!seen_children) {
		// the task is waiting for a child to die but it hasnt got any children.
		tp->tss.eax = 0xffffffff;
		add_to_ready_list(tp);
	}
}





void
drop_from_blocked_on_list(TASK_INFO * t,TASK_INFO * to_remove)

{
#if 1
	int i;

	for (i=0;i<TASK_MAX;i++) {
		if (TASK_ALIVE(&task_info[i])) {
			if (task_info[i].tasks_send_blocked == to_remove) {
				task_info[i].tasks_send_blocked = to_remove->next_send_blocked;
			}
			if (task_info[i].next_send_blocked == to_remove) {
				task_info[i].next_send_blocked = to_remove->next_send_blocked;
			}
		}
	}
#else
	// this strategy does not appear to work correctly. why ?

	// if this task is the first task in the list put the next one at front

	if (t->tasks_send_blocked == to_remove) {
		t->tasks_send_blocked = to_remove->next_send_blocked;
		return;
	}
	
	t = (TASK_INFO *) t->tasks_send_blocked;
	while (t && t->next_send_blocked && t->next_send_blocked != to_remove) {
		t = (TASK_INFO *) t->next_send_blocked;
	}

	if (t && t->next_send_blocked == to_remove) {
		t->next_send_blocked = (TASK_INFO *) to_remove->next_send_blocked;
	}
#endif
}


void
sys_exit(TASK_INFO * tp,int status,int transforming)

{
	// the task is unofficially dead. 
	// free all the space and tell everyone but keep the
	// slot so the parent may do a 'wait'

	int i;

TRACE;

	// 
	// detach the dying task from any ports & interrupts it may be attached to
	//

TRACE;
	for (i=0;i<PORT_MAX;i++) {
		if (ports[i].tid == tp->tid) {
			ports[i].tid = -1;
			int j;
			for (j=0;j<256;j++) {
				if (attached_interrupts[j] == tp->tid) {
					init_gate(&idt[j],TASK_CSEL,(word32)ivec0+stubsize*j,TYPE_INTERRUPT_GATE | TYPE_DPL_0);
					attached_interrupts[j] = -1;
					attached_igates[j] = FALSE;
				}
			}
		}
	}
TRACE;

	tp->exit_status = status;
	if (tp->parent_tid == 0) {
		// task was an orphan
		tp->state = TASK_STATE_FREE;
	}
	else {
		tp->state = TASK_STATE_ZOMBIE;
	}
TRACE;

	// check here if a task is a boot-time loaded task.
	// if it is then pmem_release its memory rather than freeing it
	// because it was allocated using pmem_claim.

	// WERK
	// if etc...
TRACE;

#if 0
	vmem_free((word32)tp->taddr);
	vmem_free((word32)tp->daddr);
#endif

	//
	// loop on all tasks to see which ones are affected by this sad occasion
	//

	for (i=0;i<TASK_MAX;i++) {
		if (TASK_ALIVE(&task_info[i])) {
			//
			// unblock any tasks blocked on the dying one
			//
			if (task_info[i].blocked_on == tp->tid) {
				task_info[i].reply_locked = FALSE;
				task_info[i].tss.eax = 0xffffffff;
				add_to_ready_list(&task_info[i]);
			}
			//
			// orphanize running children
			//
			if (!transforming) {
				if (task_info[i].parent_tid == tp->tid) {
					task_info[i].parent_tid = 0;
				}
			}
		}
		else {
			//
			// free zombie children
			// (for sure they wont be wait-ed on)
			//
			if (task_info[i].state == TASK_STATE_ZOMBIE) {
				if (task_info[i].parent_tid == tp->tid) {
					task_info[i].state = TASK_STATE_FREE;
				}
			}
		}
	}
TRACE;

	//
	// free any unprocessed obituaries if this was an admin	
	//

	while (tp->obituaries) {
		drop_obituary(tp);
	}
TRACE;

	//
	// send obituaries to all admins
	// only if we arent transforming.
	// if we are we want the admins to keep the idea that we are the same
	// task as earlier so we dont tell them anything at all.
	//
	if (!transforming) {
TRACE;
		for (i=0;i<TASK_MAX;i++) {
TRACE;
			if (TASK_ALIVE(&task_info[i])) {
				if (i == tp->tid) {			
					continue;
				}
				if (task_info[i].admin) {
					OBITUARY * op;
					op = blank_obituaries;
					if (op != NULL) {
						blank_obituaries = (OBITUARY *) op->next;
						op->tid  = tp->tid;
						op->next = task_info[i].obituaries;
						task_info[i].obituaries = op;
						// attempt to wakeup the admin in case it is already receiving
TRACE;
						wakeup_recv(&task_info[i]);
TRACE;
					}
				}
			}
		}
TRACE;
	}
TRACE;

	//
	// if this task is blocked on another one remove it from the list
	// when transforming this will automatically remove us from the
	// list of tasks blocked on the loader (reply blocked in this case).
	//
	if (tp->blocked_on != -1) {
TRACE;
		drop_from_blocked_on_list(&task_info[tp->blocked_on],tp);
	}
TRACE;

	//
	// if the parent is waiting wake it up now (only if we arent transforming)
	// 
	if (!transforming) {
TRACE;
		wakeup_wait(&task_info[tp->parent_tid]);
TRACE;
		sys_kill(tp->parent_tid,SIGCHLD);
	}
TRACE;
}

void
sys_wait(TASK_INFO * tp,word32 status_pointer)

{
	tp->wait_status_pointer = status_pointer;
	tp->state = TASK_STATE_WAIT;

	wakeup_wait(tp);
}

void
signal_call_handler(TASK_INFO * tp,word32 handler)

{
	int contextsize;
	word32 esp;
	word32 magic = 0x03110320;
	word32 phoney = 0xffffffff;

	esp = tp->daddr + tp->tss.esp;

	contextsize = (char*)&tp->blocked_on - (char*)tp;

	contextsize += 3;

	contextsize &= ~3;

	esp -= contextsize;

	itmemcpy(LINM_DSEL,esp,task_info[0].daddr+(word32)tp,contextsize);

	esp -= 4;

	itmemcpy(LINM_DSEL,esp,task_info[0].daddr+(word32)&magic,sizeof(int));

	esp -= 4;

	itmemcpy(LINM_DSEL,esp,task_info[0].daddr+(word32)&phoney,sizeof(int));

	tp->tss.esp = esp - tp->daddr;
	tp->tss.eip = handler;

	switch (tp->state) {
		case TASK_STATE_SEND:
			// this task is still send blocked; remove it from the list
			drop_from_blocked_on_list(&task_info[tp->blocked_on],tp);
			break;
		case TASK_STATE_READY:
			/* this task is already ready. nothing further to do */
			return;
			break;
	}

	/* now make the task ready */

	add_to_ready_list(tp);
}

int
signal_return_handler(TASK_INFO * tp)

{
	int contextsize;
	word32 esp;
	word32 magic;
	word32 phoney;

	esp = tp->daddr + tp->tss.esp;	

	itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)&phoney,esp,sizeof(int));

	if (phoney != 0xffffffff) {
		return -1;
	}

	esp += 4;

	// fetch magic word and check it. if it isnt right this was no signal
	// return

	itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)&magic,esp,sizeof(int));
	
	if (magic != 0x03110320) {
		return -1;
	}

	esp += 4;

	contextsize = ((char*)&tp->blocked_on) - ((char*)tp);

	// restore the context. automatically restores esp so we leave it alone

	itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)tp,esp,contextsize);

	if (tp->state != TASK_STATE_READY) {
		tp->tss.eax = 0xfffffffe;	// (-2 for restart of system call)
		tp->state = TASK_STATE_READY;
	}

	return 0;
}

void
signal_pending(TASK_INFO * tp)

{
	int sig;
	int i;

	// if the task is processing its select data dont interrupt it in the middle

	if (tp->select_sig_block) {
		return;	
	}

	// if the task is presently going to be replied to leave it alone
	// (signal will come at the end of the reply)

	if (tp->reply_locked) {
		return;
	}

	// no signals pending that arent blocked then no need to search
	// unblock SIGKILL just to be sure

	tp->sig_blocked &= ~sigmask(SIGKILL);

	if ((tp->sig_pending & ~tp->sig_blocked) == 0) {
		return;
	}

	for (sig=1;sig<NSIG;sig++) {
		i = sigmask(sig);

		if ((tp->sig_pending & i) && (tp->sig_blocked & i) == 0) {
			// signal is pending and not blocked
			switch (tp->handlers[sig]) {
				case 0:
					// default action
					switch (sigdefaults[sig]) {
						case CORE:
						case KILL:
							sys_exit(tp,sig,NOT_TRANSFORMING);
							break;
						case DISCARD:
							break;				
					}
					break;
				case -1:
					// ignore
					break;
				default:
					// call a handler
					signal_call_handler(tp,tp->handlers[sig]);
					// block the signal as long as the handler runs
					// (context restore will restore the mask)
					tp->sig_blocked |= i;
					break;
			}
			// reset the bit in the pending mask to avoid doing this over and
			// over again
            tp->sig_pending &= ~i;
			// reset the signal handler to default
			tp->handlers[sig] = 0;
			// and exit (lowest signal number is most likely the most urgent)
			// and two signals at the same time shouldnt occur
			return;
		}
	}
}


//
// set the signal bits
//

int
sys_kill(int tid,int sig)

{
	if (tid <= 0 || tid >= TASK_MAX) {
		return -1;
	}

	if (!TASK_ALIVE(&task_info[tid])) {
		// you cant kill whats already dead.
		return -1;
	}

	if (sig == SIGKILL) {
		task_info[tid].reply_locked = FALSE;
		task_info[tid].select_sig_block = FALSE;
	}

	task_info[tid].sig_pending |= sigmask(sig);

	signal_pending(&task_info[tid]);

	return 0;
}

int
find_free_task(void)

{
	static int start_from = 0;
	int i;

	//
	// this should be a list of free task slots
	//
	i = start_from;
	do {
		if (task_info[i].state == TASK_STATE_FREE) {
			// make sure next time the searching will start from the next slot.
			// this marginally increases the speed of adding a new task and it
			// also ensures that task id's wont duplicate too fast.
			start_from = i + 1;
			if (start_from == TASK_MAX) {
				// start from one because task id 0 will never be reused
				start_from = 1;
			}
			return i;
		}
		i++;
		if (i == TASK_MAX) {
			// restart search from the number of boot tasks so boot task ids
			// wont be issued twice.
			i = n_boot_tasks;
		}
	} while (i != start_from);
#ifdef DEBUG
	print(10,10,"no more task slots");
#endif

	return -1;

}

//
// a task has been successfully loaded
// now add it to the taskers internal structures
//

INLINE int
add_task(char * name,int privilege,int parent,word32 taddr,word32 tlen,word32 entry,word32 daddr,word32 dlen,word32 esp,word32 edi,word32 esi,word32 ebx,word32 brk,word32 ecx,word32 ebp,int priority)

{
	int csel;
	int dsel;
	int tsel;
	int i;

TRACE;
	if (parent & TID_DEAD) {
TRACE;
		//
		// this is a transform (execve) taking place!
		// we already know the slot number.
		//
		i = parent & ~TID_DEAD;
		// re-read the real parent.
		parent = task_info[i].parent_tid;
TRACE;
		sys_exit(&task_info[i],0,TRANSFORMING);
TRACE;
	}
	else {
TRACE;
		i = find_free_task();
TRACE;

		if (i == -1) {
TRACE;
			return -1;
		}

		task_info[i].priority = priority;
	}

	// WORK there are rules for this part
	task_info[i].sig_blocked = 0;
	task_info[i].sig_pending = 0;
TRACE;

	internal_memcpy((unsigned char *)task_info[i].name,(unsigned char *)name,TASK_NAMELEN);
	task_info[i].name[TASK_NAMELEN] = 0;

	internal_memcpy(task_info[i].fp_state,dft_fp_state,sizeof(dft_fp_state));
	task_info[i].fp_stored = TRUE;
	task_info[i].fp_used = FALSE;

	task_info[i].tid = i;

	task_info[i].parent_tid = parent;

	task_info[i].tasks_send_blocked = NULL;
	task_info[i].next_send_blocked = NULL;
	task_info[i].ports_signalled    = NULL;
	task_info[i].in_list = FALSE;
	task_info[i].next_ready = NULL;
	task_info[i].admin = FALSE;
	task_info[i].obituaries = NULL;
	task_info[i].reply_locked = FALSE;

	task_info[i].taddr = taddr;
	task_info[i].tlen  = tlen;
	task_info[i].daddr = daddr;
	task_info[i].dlen  = dlen;

	task_info[i].select_sig_block = FALSE;
	task_info[i].select_blocked   = FALSE;

	// this task is not profiling
	task_info[i].profile = 0;

	system_timers.task_timers[i] = 0;

TRACE;
	if (i == 0) {
TRACE;
		// TASK has already been initialised; its where you are now
		// we just needed to get here to fill in the fields above in
		// a uniform way.
		task_info[i].priority = 0;
		task_info[i].state    = TASK_STATE_READY;

	    gdt[USER_TSEL >> 3].type_dpl |= TYPE_DPL_3;
TRACE;

		return i;
	}
TRACE;

	tsel = USER_TSEL + (GDT_SLOTS_PER_TASK * i * sizeof(DESCRIPTOR));
	csel = tsel + 8;	// LDT ! WORK
	dsel = csel + 8;
	
	// store tsel for quick reference

	task_info[i].tsel = tsel;

	init_descriptor(csel,taddr,tlen,TYPE_CODE_DESCR | privilege,SEG_DEFAULT);
	init_descriptor(dsel,daddr,dlen,TYPE_DATA_DESCR | privilege,SEG_DEFAULT);
	init_descriptor(tsel,task_info[0].daddr + (word32)&task_info[i].tss,TSS_SIZE-1,TYPE_TSS_DESCR | TYPE_DPL_3,0);

	task_info[i].brk  = brk;
TRACE;

	init_tss(&task_info[i].tss,privilege >> 5,csel,dsel,esp,i,edi,esi,ebx,entry,daddr,ecx,ebp);
TRACE;

	add_to_ready_list(&task_info[i]);
TRACE;

	//
	// set all signals to take the default action.
	//
	{
		int signum;
TRACE;
		for (signum=1;signum<NSIG;signum++) {
			// WORK should be only with tasks not forked or execvd
			task_info[i].handlers[signum] = 0;
		}
TRACE;
	}

#ifdef DEBUG
	print(10,10+i,"allocated task in slot ");
#endif
TRACE;
	return i;
}

int
sys_fork(int parent,word32 taddr,word32 daddr)

{
	int csel;
	int dsel;
	int tsel;
	int tid;

	tid = find_free_task();

	if (tid == -1) {
		return -1;
	}

	internal_memcpy((char*)&task_info[tid],(char*)&task_info[parent],sizeof(TASK_INFO));
	task_info[tid].tid          = tid;
	task_info[tid].taddr        = taddr;
	task_info[tid].daddr        = daddr;
	task_info[tid].parent_tid   = parent;

	task_info[tid].obituaries = NULL;
	task_info[tid].tasks_send_blocked = NULL;
	task_info[tid].ports_signalled = NULL;

	tsel = USER_TSEL + (GDT_SLOTS_PER_TASK * tid * sizeof(DESCRIPTOR));
	csel = tsel + 8;	// LDT ! WORK
	dsel = csel + 8;
	
	// store tsel for quick reference

	task_info[tid].tsel = tsel;

	init_descriptor(csel,taddr,task_info[tid].tlen,TYPE_CODE_DESCR | TYPE_DPL_3,SEG_DEFAULT);
	init_descriptor(dsel,daddr,task_info[tid].dlen,TYPE_DATA_DESCR | TYPE_DPL_3,SEG_DEFAULT);
	init_descriptor(tsel,task_info[0].daddr + (word32)&task_info[tid].tss,TSS_SIZE-1,TYPE_TSS_DESCR | TYPE_DPL_3,0);

	/* WERK task struct aanpassen met privilige level ! */

	task_info[tid].tss.cs = csel | 3;
	task_info[tid].tss.ds = dsel | 3;
	task_info[tid].tss.es = dsel | 3;
	task_info[tid].tss.ss = dsel | 3;
	task_info[tid].tss.fs = dsel | 3;
	task_info[tid].tss.gs = dsel | 3;

	return tid;
}

int
sys_signal(TASK_INFO * tp,int signum,word32 handler)

{
	/*
 	 * you cant change SIGKILL and the number should be a valid
	 * signal number.
	 */
	if (signum == SIGKILL || signum < 1 || signum >= NSIG) {
		return -1;
	}

	tp->handlers[signum] = handler;

	return 0;
}

//
// sys_setcseg is used by the loader to lock; inquire and set code segment
// re-use information. cseg is the number of the segment; vbase is the 
// pointer to the base address and nusers is the current number of users
// the loader thinks there should be.
// the life cycle of a cseg is something like this:
//    on startup a cseg is initialised to all 0s. then 
//    then on first use (vbase != 0) it will just be set; nusers will be
//       2 right away so the tasker wont be tempted to free it
//    whenever new tasks are added or disappear 
//       the loader updates the nusers value (vbase still != 0)
//    there will come a time when nusers reaches 1. the tasker is now
//       free to re-use this memory should it wish so. if the loader
//       gets another client for this code segment it will ATTEMPT to 
//       set the number of users from one to two and request the vbase
//       this will return NULL if the code segment was re-used otherwise
//       the code segments user count will be set to 2 and the vbase address
//       will be returned.
//

int
sys_setcseg(int cseg,int vbase,int nusers)

{
	if (vbase) {
		// this is not an inquiry; just a set
		code_segments[cseg].nusers = nusers;
		code_segments[cseg].vbase  = vbase;
		return code_segments[cseg].vbase;
	}

	if (code_segments[cseg].nusers == 1) {
		code_segments[cseg].nusers = nusers;
		return code_segments[cseg].vbase;
	}

	return 0;
}

//
// lock or unlock a task for signals. this may only happen if the task is
// reply blocked on the caller.
// this mechanism is used to allow for signals during reads from slow devices.
// it is the device administrators responsibility to unlock a task it sees it
// has no data for. when the data arrives first the task is locked again. if
// this fails there will be no reply and the data remains in the buffer. if
// the lock succeeds it means the task is still hanging reply blocked and the
// reply takes place. the reply implicitly unlocks the task.
//

word32
sys_reply_lock(TASK_INFO * tp,int tid)

{
	if (tid < 0 && tid >= TASK_MAX) {
		return 0xffffffff;
	}

	if (task_info[tid].blocked_on == tp->tid && task_info[tid].state == TASK_STATE_REPLY) {
		task_info[tid].reply_locked = TRUE;
		return 0;
	}

	return 0xffffffff;
}

word32
sys_reply_unlock(TASK_INFO * tp,int tid)

{
	if (tid < 0 && tid >= TASK_MAX) {
		return 0xffffffff;
	}

	if (task_info[tid].blocked_on == tp->tid && task_info[tid].state == TASK_STATE_REPLY) {
		task_info[tid].reply_locked = FALSE;

		signal_pending(&task_info[tid]);
		return 0;
	}

	return 0xffffffff;
}

void
remove_port_from_list(TASK_INFO * tp,PORT * p)

{
	PORT * new_list;

	if (tp->ports_signalled == p) {
		tp->ports_signalled = (PORT * )p->next;
	}
	else {
		new_list = tp->ports_signalled;

		while (new_list->next) {
			if ((PORT *)new_list->next == p) {
				new_list->next = p->next;
			}
			else {
				new_list = (PORT *)new_list->next;
			}
		}
	}
}

INLINE
int
sys_read_port(TASK_INFO * tp,int port)

{
	int v = -1;

	if (port > 0 && port < PORT_MAX) {
		if (ports[port].tid == -1) {
			return ERROR;
		}														// nobody attached to this port; this must be an error

		v = ports[port].val;

		if (ports[port].tid == tp->tid) {
			remove_port_from_list(tp,&ports[port]);
			ports[port].val = 0;										// increment semaphore instead of simply setting to avoid losing interrupts
		}
	}
	return v;
}

void
sys_select_start(TASK_INFO * tp,int ipcunblock)

{
	tp->select_sig_block = TRUE;
	tp->select_blocked = TRUE;
	tp->tss.eax = 0;
}

void
sys_select(TASK_INFO * tp,int ipcunblock)

{
	if (tp->select_blocked) {
		tp->state = TASK_STATE_SELECT;
		tp->select_sig_block = FALSE;
		tp->ipcunblock = ipcunblock;
		wakeup_recv(tp);
	}
}

void
sys_select_end(TASK_INFO * tp)

{
	tp->select_sig_block = FALSE;
	tp->select_blocked = FALSE;
}

void
sys_select_unblock(int tid,int why)

{
	TASK_INFO * tp;

	if (tid >= 0 && tid < TASK_MAX) {
		tp = &task_info[tid];
		if (tp->state == TASK_STATE_SELECT) {
			tp->select_sig_block = TRUE;
			add_to_ready_list(tp);
			tp->tss.eax = why;
		}
		// free the task from becoming select blocked. (if it was this is harmless)
		tp->select_blocked = FALSE;
	}
}

//
// map in a part of device memory
// into a user task
//

word32
sys_mapdev(TASK_INFO  * tp,int a,int l)

{
	word32 v;
	int n;

	n = (l + (4096 - 1)) >> 12;
	n <<= 12;							// number of bytes rounded up to pagesize

	if (a < 0xa0000) {
		return 0;
	}

	if ((a + l) > 0x100000) {
		return 0;
	}

	// round the break to a whole page

	tp->brk += 4095;

	tp->brk &= ~4095;

	v = tp->brk;

	tp->brk += n;

	set_page_table_entries((void*)(v + tp->daddr),(void*)a,l);

	return v;
}


void
init_page_fault_handler(void)

{
	init_descriptor(PFLT_TSEL,task_info[0].daddr + (word32)&page_fault_handler_tss,TSS_SIZE-1,TYPE_TSS_DESCR | TYPE_DPL_3,0);

	internal_memcpy((unsigned char *)&page_fault_handler_tss,(unsigned char *)&boot_tss[0],sizeof(TSS));

	page_fault_handler_tss.eip     = (word32)page_fault_handler;
	page_fault_handler_tss.eflags &= 0xffffbdffl;
	page_fault_handler_tss.esp     = (word32) page_fault_handler_stack + 1024;
	page_fault_handler_tss.esp    &= 0xfffffff0;

	init_gate(&idt[14],PFLT_TSEL,0,TYPE_TASK_GATE | TYPE_DPL_0);
}

//
// set the task state table and ready pointers
// to known harmless values
//


void
init()

{
	int i;
	BOOT_TASK * bt;
	word32 daddr;
	word32 taddr;
	word32 tlen;
	word32 brk;
	word32 esp;
	int stacksize;

	hw_port_signalled = 0;

	// find out where the tasker lives

	__daddr = boot_tasks[0].daddr;

	// initialise the page tables and the virtual memory managers
	// data structures.

	cr3 = vmem_init();

	// FROM HERE ONWARD PAGING IS ENABLED !

	set_interrupt_stacks(&boot_tss[0]);

	fpsave((char *)dft_fp_state);

	/*
	 * since these set the idt to point to a real mode code segment
	 * the handler should be in mode.asm and not in the gnu compiled part
	 */
	stubsize = (word32) ivec1 - (word32) ivec0;

	for (i=0;i<256;i++) {
		init_gate(&idt[i],TASK_CSEL,(word32)ivec0+stubsize*i,TYPE_INTERRUPT_GATE | TYPE_DPL_0);
		attached_interrupts[i] = -1;
		attached_igates[i] = FALSE;
	}

	// setup a segment for the interrupt stacks parallel to taskers data segment

	gdt[INTS_DSEL >> 3] = gdt[TASK_DSEL >> 3];

	init_gate(&idt[7],TASK_CSEL,(word32)fprestore,TYPE_INTERRUPT_GATE | TYPE_DPL_0);
	init_gate(&idt[0x78],TASK_CSEL,(word32)timer_interrupt,TYPE_INTERRUPT_GATE | TYPE_DPL_0);
	init_gate(&idt[254],TASK_CSEL,(word32)nosignal_hw_port,TYPE_INTERRUPT_GATE | TYPE_DPL_0);
	init_gate(&idt[255],TASK_CSEL,(word32)signal_hw_port,TYPE_INTERRUPT_GATE | TYPE_DPL_0);

	for (i=0;i<TASK_MAX;i++) {
		internal_memset((unsigned char *)&task_info[i],0,sizeof(TASK_INFO));
		task_info[i].tid   = i;
		task_info[i].state = TASK_STATE_FREE;
 		task_info[i].tasks_send_blocked = NULL;
 		task_info[i].next_send_blocked = NULL;
		task_info[i].ports_signalled    = NULL;
		task_info[i].obituaries = NULL;
		task_info[i].in_list = FALSE;
		task_info[i].next_ready = NULL;

		code_segments[i].nusers = 0;
		code_segments[i].vbase  = 0;
	}

	for (i=1;i<PORT_MAX;i++) {
		ports[i].tid = -1;
		ports[i].val = 0;
		ports[i].next = NULL;
		ports[i].num = i;
	}

	for (i=0;i<PRI_MAX;i++) {
		ready_tasks[i] = NULL;
	}

	current_task = NULL;

	//
	// initialise the obituary list
	//

	internal_memset((unsigned char *) obituaries,0,sizeof(obituaries));

	for (i=0;i<MAX_OBITUARIES-1;i++) {
		obituaries[i].next = &obituaries[i+1];
	}

	blank_obituaries = obituaries;

/*	init_descriptor(TTSS_DSEL,task_info[0].daddr + (word32)&boot_tss[0].tss,TSS_SIZE-1,TYPE_DATA_DESCR | TYPE_DPL_3,0);
*/
	//
	// set up the tasks loaded by the boot loader
	//
	for (i=0;i<n_boot_tasks;i++) {
		bt = &boot_tasks[i];
#ifdef DEBUG
		print(10,i+5,bt->name);
#endif
		daddr = bt->daddr;
		taddr = bt->taddr;
		brk   = bt->header.bsize + bt->header.dsize;
		tlen  = bt->header.tsize + sizeof(bt->header) + 4096;
		// the bt->argv-4 is the address of the beginning of the argv array;
		// it is the last data the loader pushed on the stack. below this point
		// the stack is free to be used by the program

		esp = bt->argv - 4;

		if (i == 0) {
			add_task(bt->name,TYPE_DPL_0,0,taddr,tlen,bt->header.entry,daddr,bt->header.dlimit,esp,bt->argc,bt->argv,bt->envp,brk,0,0,bt->priority);

			task_info[0].tss.cr3 = cr3;
			boot_tss[0].cr3 = cr3;
			boot_tss[1].cr3 = cr3;
		}
		else {
			// claim the pages the task uses to store its text

			pmem_claim((word32) bt->taddr,tlen);

			if (bt->header.info & 0x80000000) {
				//
			 	// task is not virtualizable; (i.e. because it uses DMA)
				// just claim its memory and leave it where it is
				// there is no need to map anything because all physical
				// memory is already mapped unto itsself
				//

				pmem_claim((word32)bt->daddr,bt->header.dlimit);
			}
			else {
				//
				// task is virtualizable.
				//

				// allocate space for the text segment in virtual memory

				taddr = (word32) vmem_alloc(tlen);

				// and remap the (already filled and claimed) text pages
				// to the virtual ones

				set_page_table_entries((void*)taddr,(void*)bt->taddr,tlen);

				// allocate space for the data segment in virtual memory

				daddr = (word32) vmem_alloc(bt->header.dlimit);
#if 1
				// claim the data space

				pmem_claim((word32)bt->daddr,brk);

				// remap the dss and bss to virtual memory

				set_page_table_entries((void*)daddr,(void*)bt->daddr,brk);

				// find out how much stack to use

				stacksize = bt->header.dlimit - esp;

				if (stacksize < 4096) {
					stacksize = 4096;
				}

				if ((word32) stacksize > bt->header.dlimit) {
					stacksize = bt->header.dlimit;
				}

				// and stack space

				pmem_claim((word32)(bt->daddr+bt->header.dlimit) - stacksize,stacksize);

				// remap the stack to virtual memory

				set_page_table_entries((void*)((daddr + bt->header.dlimit) - stacksize),(void*)((bt->daddr+bt->header.dlimit)-stacksize),stacksize);
#else
				pmem_claim((word32)bt->daddr,bt->header.dlimit);
				set_page_table_entries((void*)daddr,(void*)bt->daddr,bt->header.dlimit);
#endif
			}
			add_task(bt->name,TYPE_DPL_3,0,taddr,tlen,bt->header.entry,daddr,bt->header.dlimit,esp,bt->argc,bt->argv,bt->envp,brk,0,0,bt->priority);
		}
	}

	outportb(TIMER_MODE,SQUARE_WAVE);
	outportb(TIMER0,TIMER_COUNT);
	outportb(TIMER0,TIMER_COUNT >> 8);

	vmem_wipe();

	init_descriptor(STIM_DSEL,task_info[0].daddr + (word32)&system_timers,sizeof(system_timers)-1,TYPE_DATA_DESCR | TYPE_DPL_3,0);

	init_page_fault_handler();

}

INLINE
void
check_interrupts(void)

{
	if (int_number == 0x78) {
		(void) sys_signal_port(PORT_TIMER);
		int_number = 0x100;
	}

	//
	// check if any hw_signal_ports were done. if so signal the ports
	// involved.
	//
	while (hw_port_signalled) {
		hw_port_signalled--;
		if (sys_signal_port(hw_ports_signalled[hw_port_signalled]) !=
			hw_ports_signalled[hw_port_signalled]) {
			print(5,5,"***** can't signal interrupt port *****");
			print_num(5,6,hw_ports_signalled[hw_port_signalled]);
		}
	}
}

//
// restore the processor context for the task pointed to by tp
// and return; in effect resuming the task where it last left off
//

extern "C" {
	void jump_to_task(int tsel);
};

//
// changeover to a user task.
// set the current task pointer; enable the interrupts for the task
// being started and jump to it. on return evaluate the reason for the
// return (software interrupt/hardware interrupt/trap)
// and act accordingly
//

// which faults have an error pushed on the stack

int has_error[18] = {
	0,0,0,0,0,0,0,0,1,0,1,1,1,1,1,0,0,0
};

// how many bytes of parameters 
// with every system call

int callparameters[] = {
	0,20,12,12,4,0,4,4,12,4,0,4,4,20,4,4,60,0,4,4,8,8,8,0,12,8,0,0,0,4,4,12,0,12,0,4,4,4,4,4,0,4,8,4,4,32
};

// the parameters for this system call

int parameters[20];

word32 tempregs[3];

char name[32];

word32
fault_address(void)

{
	asm("	movl	%cr3,%eax");
}

void
mark_tasker_busy(int ds)

{
	if (ds != -1) {
		hardware_interrupt_tss[(ds >> 3) - 32 - (VECT_GSEL >> 3) - 0x70].back_link = 0x100;
	}
	else {
		page_fault_handler_tss.back_link = 0x100;
	}
	gdt[TASK_TSEL >> 3].type_dpl |= 0x02;
}

//
// find out how long it has been since the last call to took took place
// by reading the hardware timer. 
// there may be a problem here reading the lsb and getting a changed msb
//

INLINE
int
took(void)

{
	REGISTER int t;
	REGISTER int v;
	static int prev_timerval = 0;

	v  = inportb(TIMER0);
	v |= inportb(TIMER0) << 8;

	if (v > prev_timerval) {
		// weve had a reload. the time we werent here is
		t = prev_timerval + (TIMER_COUNT - v);
	}
	else {
		// we didnt have a reload. 
		t = prev_timerval - v;
	}

	prev_timerval = v;

	return t;
}

void
increment_profile_counter(TASK_INFO * tp)

{
	int bucket;
	short int count;

	// taddrlow = 0x1020; task entry address

	bucket = tp->tss.eip - 0x1020;

	// one bucket every four bytes

	bucket >>= 2;

	itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)&count,tp->daddr + tp->profile + (bucket << 1),2);

	count++;

	itmemcpy(LINM_DSEL,tp->daddr + tp->profile + (bucket << 1),task_info[0].daddr+(word32)&count,2);
}

//
// context switch to a new task entered with interrupts disabled because
// no more checks are done until we are in a user process so any interrupts
// coming between now and then would be lost.
//

INLINE void
context_switch(REGISTER TASK_INFO * tp)

{
	int n;
	word32 esp;
	int local_int_number;

#ifdef VISUAL_STATES
	if (tp->tid < 80) {
		vid_mem_spinner(tp->tid,23);
	}
#endif
	//
	// remember which is the current task. useful global for system calls
	//
	current_task = tp;

	//
	// bad programming !	
	// this task is not ready so it shouldn't be in the list. 
	// probably it exited with a signal or something like that and was still in
	// the ready list. handle the situation here. that's much easier and more
	// streamlined than forcefully removing the task from the lists.
	//
	if (current_task->state != TASK_STATE_READY) {
		return;
	}

	//
	// reset the task busy bit
	//
	tp->tss.eflags &= 0xffffbdffl;
	tp->tss.back_link = 0;

	gdt[tp->tsel >> 3].type_dpl &= ~SEG_TASK_BUSY;

	_disable();

	gdt[TASK_TSEL >> 3].type_dpl &= ~SEG_TASK_BUSY;
	gdt[(VECT_GSEL >> 3) + 0x79].type_dpl &= ~SEG_TASK_BUSY;

	//
	// enable interrupts
	// and set iopl to 3
	//

	tp->tss.eflags |= 0x00003200l;

	fp_used = &tp->fp_used;
	if (tp->fp_used) {
		fp_restore((char *)tp->fp_state);
	}

	// non existing vector number to catch system calls

	int_number = 0x100;

	//
	// log how much time we spent in the tasker
	//

	n = took();

	system_timers.task_timers[0]     += n;
	system_timers.priority_timers[0] += n;

asm("	movl	$0x378,%edx   ");
asm("	movl	$0xff,%eax    ");

asm("	.byte	0xee          ");
	//
	// and jump to the task
	//

#ifdef VISUAL_STATES
	vid_mem_putchar(tp->tid,22,'v');
#endif

	jump_to_task(tp->tsel);

#ifdef VISUAL_STATES
	vid_mem_putchar(tp->tid,22,'*');
#endif

asm("	movl	$0x378,%edx   ");
asm("	movl	$0x00,%eax    ");

asm("	.byte	0xee          ");

	//
	// log how much time we spent in the tasker
	//

	n = took();

	system_timers.task_timers[tp->tid]          += n;
	system_timers.priority_timers[tp->priority] += n;

	//
	// back from the task. copy the interrupt number why were back 
	// to something local and re-enable interrupts
	//
	local_int_number = int_number;

#if 0
	if (local_int_number != 0x78 && local_int_number != 0x100) {
		if (local_int_number < 0x20) {
			print_num(60,19,local_int_number);
		}
		else {
			if (local_int_number == 0xfe) {
				print_num(60,22,local_int_number);
			}
			else {
				print_num(60,21,local_int_number);
			}
		}
	}
	else {
		print_num(60,20,local_int_number);
	}
#endif

	if (tp->fp_used) {
		fpsave((char *)tp->fp_state);
	}

	if (local_int_number == 0xfe) {
		return;
	}

	//
	// the current task requested some more memory but none was available
	// try to free up some memory. task restart will clean up this mess
	// maybe move it to pager task but thats risky
	//
	if (local_int_number == 14) {
		if (sys_free_old_code_segment() < 0) {
			// should be something like sigstack
			sys_kill(tp->tid,SIGSEGV);
		}
		return;
	}

	//
	// back from the user task. it did a int xx; made a mistake
	// or was interrupted by a piece of hardware.
	// now interrupts are disabled again.
	//

	// get the lineair address pointing to the stack

	// WORK should validate the stack pointer here before referencing
	// if the task got messed up it may not be valid.

	esp = tp->daddr + tp->tss.esp;

	if (local_int_number == 0x100) {
		_enable();

#if 0
		print(0,1,"task          call      ");
		print_num(5,1,tp->tid);
		print_num(21,1,tp->tss.eax);
#endif
		//
		// this wasnt a timeout
		// so it must have been a system call. find out which one
		// and act accordingly.
		//

		//
		// since a system call can take up a lot of time we enable interrupts
		// before calling again to context switch the interrupts will be checked
		//
#ifdef DEBUG
		vid_mem_putchar(18,20+current_task->tid,'0'+current_task->tid);

		vid_mem_putchar(20,20+current_task->tid,current_task->tss.eax + '0');
#endif
		if (tp->tss.eax < 1 || tp->tss.eax > SYSCALLS_LAST) {
			// this system call doesn't exist   
			// so kill the task
			core_dump();
			return;
		}

		//
		// WORK check if there are enough parameters on the stack here
		// should be; if someone came in through a library call
		//
		n = callparameters[tp->tss.eax];
		if (n) {
			itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)&parameters,esp+4,n);
		}
#ifdef DEBUG
		dump_bytes(22,20+current_task->tid,(unsigned char *)parameters,callparameters[current_task->tss.eax]);

		for (i=0;i<10000000;i++) {
			nothing();
		}
#endif

		switch (tp->tss.eax) {
			case SYSCALLS_SENDM:
				_enable();
				sys_sendm(tp,parameters[0],parameters[1],parameters[2],parameters[3],parameters[4]);
				return;
				break;
			case SYSCALLS_RECEIVEM:
				_enable();
				sys_receivem(tp,parameters[0],parameters[1],parameters[2]);
				return;
				break;
			case SYSCALLS_REPLYM:
	        	_enable();
				sys_replym(tp,parameters[0],parameters[1],parameters[2]);
				return;
				break;
			case SYSCALLS_PORT_ATTACH:
				_enable();
				sys_port_attach(tp,parameters[0]);
				return;
				break;
			case SYSCALLS_TIMESLICE:
				_enable();
				return;
				break;
			case SYSCALLS_PORT_DETACH:
				_enable();
				sys_port_detach(tp,parameters[0]);
				return;
				break;
			case SYSCALLS_FORCE_READY:
	    		_enable();
				sys_force_ready(tp,parameters[0]);
				return;
				break;
			case SYSCALLS_INTERRUPT_ATTACH:
				_disable();
				sys_interrupt_attach(tp,parameters[0],parameters[1],parameters[2]);
				return;
				break;
			case SYSCALLS_SBRK:
				_enable();
				sys_sbrk(tp,parameters[0]);
				return;
				break;
			case SYSCALLS_PPAGES:
				_enable();
				tp->tss.eax = pmem_pages();
				return;
				break;
			case SYSCALLS_PMSTAT:
#if 0
				_enable();
				itmemcpy(LINM_DSEL,tp->daddr + parameters[0],task_info[0].daddr + (word32)pmem_stat(),pmem_pages() * sizeof(word32));
#endif
				return;
				break;
			case SYSCALLS_EXIT:
				// the left shift of the exit status is done here so 
				// no one will notice sys_exit is also called when a signal
				// causes the 'exit' of a task.
				// in that case the existatus will be filled with the signal number
				_enable();
				sys_exit(tp,(parameters[0] & 0xff) << 8,NOT_TRANSFORMING);
				return;
				break;
			case SYSCALLS_SET_GDT:
				_disable();
				init_descriptor((int)parameters[0],(word32)parameters[1],
								(word32)parameters[2],(word8)parameters[3],
								(word16)parameters[4]);
				return;
				break;
			case SYSCALLS_PMEM_ALLOC:
				_enable();
				tp->tss.eax = (word32) vmem_alloc(parameters[0]);
				return;
				break;
			case SYSCALLS_PMEM_FREE:
				_enable();
				tp->tss.eax = 0;
				vmem_free((word32) parameters[0]);
				return;
				break;
			case SYSCALLS_ADD_TASK:
				// get the name string from user space
				// since the loader is a 'trusted' process in the sence that
				// it is assumed not to make mistakes there is no use in 
				// checking its parameters here. we should verify this IS the
				// loader
				_enable();
				itmemcpy(LINM_DSEL,task_info[0].daddr+(word32)name,tp->daddr + parameters[0],32);
				name[32] = 0;
				tp->tss.eax = add_task(name,TYPE_DPL_3,
					// parent     taddr         tlen          entry
					parameters[1],parameters[2],parameters[3],parameters[4],
	             	// daddr      dlen          esp           edi
					parameters[5],parameters[6],parameters[7],parameters[8],
					// esi        ebx            brk            ecx
					parameters[9],parameters[10],parameters[11],parameters[12],
					// ebp        priority
					parameters[13],parameters[14]);
				return;
				break;
			case SYSCALLS_ADMIN:
				_enable();
				tp->admin = TRUE;
				return;
				break;
			case SYSCALLS_SIGNAL_PORT:
				_enable();
				tp->tss.eax = sys_signal_port(parameters[0]);
				return;
				break;
			case SYSCALLS_WAIT:
				_enable();
				sys_wait(tp,parameters[0]);
				return;
				break;
			case SYSCALLS_KILL:
				_enable();
				tp->tss.eax = sys_kill(parameters[0],parameters[1]);
				return;
				break;
			case SYSCALLS_SIGNAL:
				_enable();
				tp->tss.eax = sys_signal(tp,parameters[0],parameters[1]);
				return;
				break;
			case SYSCALLS_GET_TASK_INFO:
				_enable();
				if (parameters[0] >= 0 && parameters[0] < TASK_MAX) {
					itmemcpy(LINM_DSEL,tp->daddr + parameters[1],task_info[0].daddr + (word32)&task_info[parameters[0]],sizeof(TASK_INFO));
					tp->tss.eax = 0;
				}
				else {
					tp->tss.eax = 0xffffffff;
				}
				return;			
				break;
			case SYSCALLS_IRET_NO_SIGNAL:
				//
				// should never occur during normal processing
				//
				return;
				break;
			case SYSCALLS_IGATE_ATTACH:
				_disable();
				sys_igate_attach(tp,parameters[0],parameters[1],parameters[2]);
				return;
				break;
			case SYSCALLS_VMEM_MAP:
				_enable();
				tp->tss.eax = (word32) vmem_map((void*)parameters[0],parameters[1]);
				return;
				break;
			case SYSCALLS_PAGE_DIR:
				_enable();
				tp->tss.eax = page_dir_address();
				return;
				break;
			case SYSCALLS_PAGE_BITMAP:
				_enable();
				tp->tss.eax = pmem_bitmap();
				return;
				break;
			case SYSCALLS_REBOOT:
				_disable();
				asm("	.byte	0xf4");
				break;
			case SYSCALLS_PROFILE:
				_enable();
				tp->profile = parameters[0];
				return;
				break;
			case SYSCALLS_ISMAPPED:
				_enable();
				tp->tss.eax = pmem_ismapped(parameters[0]);
				return;
				break;
			case SYSCALLS_FORK:
				_enable();
				tp->tss.eax = sys_fork(parameters[0],parameters[1],parameters[2]);
				return;
			case SYSCALLS_GETPID:
				_enable();
				tp->tss.eax = tp->tid;
				return;
				break;
			case SYSCALLS_SETCSEG:
				_disable();
				tp->tss.eax = sys_setcseg(parameters[0],parameters[1],parameters[2]);
				return;
				break;
			case SYSCALLS_FREE_OLDCSEG:
				_disable();
				tp->tss.eax = sys_free_old_code_segment();
				return;
				break;
			case SYSCALLS_REPLY_LOCK:
				_disable();
				tp->tss.eax = sys_reply_lock(tp,parameters[0]);
				break;
			case SYSCALLS_REPLY_UNLOCK:
				_disable();
				tp->tss.eax = sys_reply_unlock(tp,parameters[0]);
				break;
			case SYSCALLS_READ_PORT:
				_enable();
				tp->tss.eax = sys_read_port(tp,parameters[0]);
				return;
				break;
			case SYSCALLS_SELECT_START:
				_disable();
				sys_select_start(tp,parameters[0]);
				return;
				break;
			case SYSCALLS_SELECT:
				_enable();
				sys_select(tp,parameters[0]);
				return;
				break;
			case SYSCALLS_SELECT_END:
				_enable();
				sys_select_end(tp);
				return;
				break;
			case SYSCALLS_SELECT_UNBLOCK:
				_disable();
				sys_select_unblock(parameters[0],0);
				return;
				break;
			case SYSCALLS_MAPDEV:
				_disable();
				tp->tss.eax = sys_mapdev(tp,parameters[0],parameters[1]);
				return;
				break;
			case SYSCALLS_SETPPID:
				_disable();
				// I dont see a good reason to allow setppid to pids other than 0
				if (parameters[0] == 0) {
					tp->parent_tid = parameters[0];
					tp->tss.eax = 0;
				}
				else {
					tp->tss.eax = 0xffffffff;
				}
				return;
				break;
			case SYSCALLS_SETPRIORITY:
				_enable();
				if (parameters[0] > 0 && parameters[0] < 16) {
					tp->priority = parameters[0];
					tp->tss.eax = 0;
				}
				else {
					tp->tss.eax = 0xffffffff;
				}
				return;
				break;
			case SYSCALLS_INT86:
				// this bit is the only reason the loader is kept in memory
				_disable();

				// all the other values are returned in registers

				boot_tss[MAIN_TSS_ID].eax = parameters[1];
				boot_tss[MAIN_TSS_ID].ebx = parameters[2];
				boot_tss[MAIN_TSS_ID].ecx = parameters[3];
				boot_tss[MAIN_TSS_ID].edx = parameters[4];
				boot_tss[MAIN_TSS_ID].esi = parameters[5];
				boot_tss[MAIN_TSS_ID].edi = parameters[6];
				// the interrupt number sits in ebp 

				boot_tss[MAIN_TSS_ID].ebp = parameters[0];
				jump_to_task(MAIN_TSEL);
				tp->tss.eax = boot_tss[MAIN_TSS_ID].eax;
				return;
				break;
		}

		return;
	}

//	int_number &= 0xff;

//	print(5+local_int_number,5,"interrupt");
//	vid_mem_putchar(20,5,local_int_number);

	if (local_int_number != 10) {
	    tp->tss.esp    = P_ISTACK_ESP;
		tp->tss.ss     = P_ISTACK_SS;
		tp->tss.eflags = P_ISTACK_EFLAGS;
		tp->tss.cs     = P_ISTACK_CS;
		tp->tss.eip    = P_ISTACK_EIP;
		error_word     = P_ISTACK_ERROR;
	}
#if 0
	}
	else {
		tp->tss.eflags = N_ISTACK_EFLAGS;
		tp->tss.cs     = N_ISTACK_CS;
		tp->tss.eip    = N_ISTACK_EIP;
		error_word     = N_ISTACK_ERROR;
	}
#endif
	_enable();

#if 0
	if (local_int_number == 10) {
		_disable();
		dump_tss(&tp->tss);
		dump_bytes(40,0,(unsigned char *)&gdt[tp->tsel >> 3],8);
		while (1) {
			nothing();
		}
	}
#endif
	if (signal_return_handler(tp) == 0) {
		return;			
	}

	if (local_int_number <= 17) {
		sys_kill(tp->tid,SIGSEGV);
		return;
	}

	if (local_int_number != 0x78) {
		//
		// spurious or not handled interrupts.
		// acknowledge them anyway to avoid hanging the system
		//

		asm("	movl	$0x20,%edx   ");
		asm("	movl	$0x20,%eax    ");
		asm("	.byte	0xee          ");
		/*
		 * end of interrupt on slave also just to be sure
		 */
		asm("	movl	$0xa0,%edx	");
		asm("	.byte	0xee		");
	}
	else {
		if (tp->profile) {
			increment_profile_counter(tp);
		}
	}

	_enable();
}

void
slice()

{
	REGISTER TASK_INFO * tp;
	REGISTER int i;

	while (1) {
foundtask:
#ifdef VISUAL_STATES
		show_task_states();
#endif

		_disable();

		check_interrupts();										// check them to make sure none occured

		_enable();

		for (i=current_priority;i<PRI_MAX;i++) {				// check all priority levels
			if (ready_tasks[i] != NULL) {						// and stop at the first one that has a ready task
				current_priority = i;							// next time search from here

				tp = ready_tasks[i];							// find it's entry
				ready_tasks[i] = (TASK_INFO*) tp->next_ready;	// and advance the list
#ifdef READY_LIST_ROBUST
				tp->in_list = FALSE;
#endif
				context_switch(tp);								// switch context and execute the task for a slice

#ifdef VISUAL_STATES
				vid_mem_putchar(tp->tid,22,' ');
#endif
	
				//
				// if the current task remains ready after a system call
				// (or for sure if it got interrupted) add it to the tail of
				// the ready list.
				//
				if (tp->state == TASK_STATE_READY) {
					add_to_ready_list(tp);
				}
				goto foundtask;
			}
		}

		// this means real trouble. if there is no task ready it also means
		// there is very little chance of one becoming ready in future.
		// show it !

		print(0,0,"task: panic ! no ready task to execute");
		for (i=0;i<1000000;i++) {
			nothing();
		}
		print(0,0,"                                      ");
		for (i=0;i<1000000;i++) {
			nothing();
		}
	}
}

int
main()

{
	/*
	 * disable interrupts and send end-of-interrupt commands to the
*	 * controllers
	 */

	_disable();

	init();

	asm("	movl	$0x20,%edx   ");
	asm("	movl	$0x20,%eax    ");
	asm("	.byte	0xee          ");
	/*
	 * end of interrupt on slave also just to be sure
	 */
	asm("	movl	$0xa0,%edx	");
	asm("	.byte	0xee		");

	while (1) {
		slice();
	}
}


